import logging
logging.getLogger().setLevel(logging.INFO)
from .utils import *
from matplotlib.path import Path
import numpy as np
from skimage import draw, measure, io
import squidpy as sq
import matplotlib.pyplot as plt
from PIL import Image
import scanpy as sc
Image.MAX_IMAGE_PIXELS = None



def create_mask(polygon, shape):
    """Create a mask for the given shape

    Parameters
    ----------
    polygon : (N, 2) array
        Points defining the shape.
    shape : tuple of two ints
        Shape of the output mask.

    Returns
    -------
    mask : (shape[0], shape[1]) array
        Boolean mask of the given shape.
    """
    polygon = polygon.iloc[:, -2:].values
    print("polygon: \n", polygon)
    polygon = np.clip(polygon, a_min=0, a_max=None)
    print("polygon adjusted: \n", polygon)
    ## Keep the order of x and y unchanged
    rr, cc = draw.polygon(polygon[:, 0], polygon[:, 1], shape) 
    # rr, cc = draw.polygon(polygon.iloc[:, -2], polygon.iloc[:, -1], shape) 

    ## Set negative coordinate to 0
    mask = np.zeros(shape, dtype=bool)
    mask[rr, cc] = True
    return mask, polygon



def crop_img_adata(roi_path, img_path, adata_path, crop_img_path, crop_adata_path, 
                   segment=False, save=None):
    """
    Crop an image and an AnnData object based on a region of interest.

    Parameters:
    roi_path : numpy.ndarray
        A numpy array specifying the region of interest.
    img_path : str
        The path to the image file.
    adata_path : str
        The path to the AnnData file.
    crop_img_path : str
        The path where the cropped image will be saved.
    crop_adata_path : str
        The path where the cropped AnnData object will be saved.
    save: bool, optional
        Whether to save the cropped image and AnnData object to files. 
        Default is None, which means not to save.

    Returns:
    tuple
        A tuple containing the cropped image and the cropped AnnData object.
    """

    roi_coords = pd.read_csv(roi_path)
    print("ROI coordinates from napari package: \n", roi_coords)

    img = io.imread(img_path)
    print("img shape: \n", img.shape)

    ## Create a mask for the region of interest
    mask, roi_coords = create_mask(roi_coords, img.shape[:2])
    ## Find the bounding box of the region of interest
    props = measure.regionprops_table(mask.astype(int), properties=('bbox',))

    minr = props['bbox-0'][0]
    minc = props['bbox-1'][0]
    maxr = props['bbox-2'][0]
    maxc = props['bbox-3'][0]

    cropped_img = img[minr:maxr, minc:maxc]
    print("cropped_img shape: \n", cropped_img.shape)

    if save:
        io.imsave(crop_img_path, cropped_img)

    adata = sc.read_h5ad(adata_path)

    print("The adata: \n", adata)
    print("The range of original adata: \n", 
          [[adata.obsm["spatial"][:,0].min(), adata.obsm["spatial"][:,0].max()], 
           [adata.obsm["spatial"][:,1].min(), adata.obsm["spatial"][:,1].max()]])
    
    ## replace x and y of adata
    roi_yx = roi_coords[:, [1, 0]]   
    # roi_yx = np.stack([roi_coords.iloc[:, -1], roi_coords.iloc[:, -2]]).T
    adata_roi = adata[Path(roi_yx).contains_points(adata.obsm["spatial"]), :].copy()

    ## Update the 'spatial' field of the AnnData object
    # print(roi_coords)
    # print([roi_coords[0][1], roi_coords[0][0]])

    ## Update the 'spatial' field of the AnnData object
    ## if you no need segment, then it can be omitted.
    if segment:
        if roi_coords[2][0] == 0: 
            adata_roi.obsm["spatial"] = adata_roi.obsm["spatial"] - \
                                        np.array([roi_coords[0][1], 0])
        else: 
            adata_roi.obsm["spatial"] = adata_roi.obsm["spatial"] - \
                                        np.array([roi_coords[0][1], roi_coords[0][0]])
    if save:
        adata_roi.write(crop_adata_path)

    return cropped_img, adata_roi


def adata_nuclei_filter(adata_sp, img_path, whole_path, roi_path):
    coord_cell = adata_sp.uns['cell_locations']
    coord_cell = coord_cell.dropna()
    coord_cell.columns = ['pxl_row_in_fullres', 'pxl_col_in_fullres', 
                          'spot_index', 'cell_index', 'cell_nums']
    
    image = plt.imread(img_path)
    img = sq.im.ImageContainer(image)
    coord_image = pd.read_csv(whole_path)
    print("Coordinates from napari package: \n", coord_image)
    _, coord_image = create_mask(coord_image, img.shape[:2])

    ## adjust adata_sp.obsm["spatial"] using the cropped whole image coords
    if coord_image[2][0] == 0: 
        adata_sp.obsm["spatial"] = adata_sp.obsm["spatial"] + \
                                    np.array([coord_image[0][1], 0])
        coord_cell[['pxl_row_in_fullres', 'pxl_col_in_fullres']] += np.array([coord_image[0][1], 0])
    else: 
        adata_sp.obsm["spatial"] = adata_sp.obsm["spatial"] + \
                                    np.array([coord_image[0][1], coord_image[0][0]])
        coord_cell[['pxl_row_in_fullres', 'pxl_col_in_fullres']] += np.array([coord_image[0][1], coord_image[0][0]])
    print("adata_sp.obsm: spatial: \n", adata_sp.obsm["spatial"])

    ## select adata_sp.obsm["spatial"] using the cropped ROI image coords
    roi_coords = pd.read_csv(roi_path)
    print("ROI coordinates from napari package: \n", roi_coords)
    _, roi_coords = create_mask(roi_coords, img.shape[:2])
    roi_yx = roi_coords[:, [1, 0]]   
    ad_sp_crop = adata_sp[Path(roi_yx).contains_points(adata_sp.obsm["spatial"]), :].copy()

    ## filter
    ad_sp_crop.uns['cell_locations'] = (
        ad_sp_crop.uns['cell_locations']
        .loc[ad_sp_crop.uns['cell_locations'].spot_index.isin(ad_sp_crop.obs.index)]
        .reset_index(drop=True)
    )
    
    return ad_sp_crop, coord_image, coord_cell


import anndata
import pandas as pd
import numpy as np
from sklearn.neighbors import KDTree
from sklearn.neighbors import BallTree
from scipy.spatial import cKDTree
import random
import numpy as np
from tqdm import tqdm
import torch
import time
from .utils import *


#######################################################################
## 2024.9.16 LLY add some function for Train and Test
#######################################################################
def extract_test_data(data_loader):
    reduced_expression_list = []
    reduced_image_list = []
    reduced_coord_list = []
    reduced_row_list = []
    reduced_col_list = []

    for batch in tqdm(data_loader):
        reduced_expression_list.append(batch['reduced_expression'])
        reduced_image_list.append(batch['image'])
        reduced_coord_list.append(batch['spatial_coords'])  
        reduced_row_list.append(batch['array_row'])  
        reduced_col_list.append(batch['array_col']) 

    print("***** batch_size=adata.shape[0] doen't effect *****")
    input_spot_test = torch.cat(reduced_expression_list)
    print(input_spot_test.shape)
    input_image_test = torch.cat(reduced_image_list)
    print(input_image_test.shape)

    input_coord_test = reduced_coord_list
    print(len(input_coord_test))
    input_row_test = reduced_row_list
    print(len(input_row_test))
    input_col_test = reduced_col_list
    print(len(input_col_test))
    print("***** *****")
    
    print("Finished extractting test data")    
    return input_spot_test, input_image_test, input_coord_test, input_row_test, input_col_test


#######################################################################
## 2024.9.16 LLY add some function for Load between spot data
#######################################################################
def extract_test_data_image_between_spot(data_loader):
    reduced_image_list = []
    reduced_coord_list = []

    for batch in tqdm(data_loader):
        reduced_image_list.append(batch['image'])
        reduced_coord_list.append(batch['spatial_coords'])  

    print("***** batch_size=adata.shape[0] *****")
    input_image_test = torch.cat(reduced_image_list)
    print(input_image_test.shape)

    input_coord_test = reduced_coord_list
    print(len(input_coord_test))
    print("***** *****")
    
    print("Finished extractting image_between_spot data")    
    return input_image_test, input_coord_test
    

def loadBatchData(train_image_mat, train_matrix_mat, train_coors_mat, batch_size, pos_info):
    '''
    Generate batch training data   
    '''
    
    train_pos_dist = pos_info['pos dist']
    train_pos_ind = pos_info['pos ind']
    
    train_index_list = list(range(train_image_mat.shape[0]))
    random.shuffle(train_index_list)

    
    train_data_size = train_image_mat.shape[0]

    half_batch_size =  int(batch_size/2)
    batch_num = train_data_size//half_batch_size
    
    for i in range(batch_num):
        
        start = i*half_batch_size
        end = start + half_batch_size
        
        tmp_index_list =  list(train_index_list[start:end])
       
        pos_peer_index = []

        neighbor_index = np.zeros((batch_size, batch_size))
        
        count = 0
        pos_index_list = []
        for j in tmp_index_list:
             
            cur_pos_peer_index = np.copy(train_pos_ind[j])
            ## shummin           
            # random.shuffle(cur_pos_peer_index)
            # pos_index_list.append(cur_pos_peer_index[0])
            
            ## when only select itself, adjust this
            # random.shuffle(cur_pos_peer_index)
            pos_index_list.append(cur_pos_peer_index)
            
            neighbor_index[count][half_batch_size+count] = 1 
            neighbor_index[half_batch_size+count][count] = 1
 
            count += 1
     
        tmp_index_list.extend(pos_index_list)
        cur_index_list = np.asarray(tmp_index_list)
        cur_batch_mat = np.take(train_image_mat.cpu(), cur_index_list, axis=0)
        cur_matrix_mat = np.take(train_matrix_mat.cpu(), cur_index_list, axis=0)
        
        yield cur_batch_mat, cur_matrix_mat, neighbor_index, cur_index_list        

    pass


#################################################################
# 2024.09.16 NameError: name 'checkNeighbors' is not defined
# 2024.11.02 Adjust parameters
#################################################################
def checkNeighbors(cur_adata, neighbor_k, tree_type='KDTree', leaf_size=2):
    '''
    parameter:
        tree_type: BallTree, cKDTree, KDTree (fast -> low)
        leaf_size: defalt 'leaf_size=2'
    Return 'dist' and 'ind' of positive samples.    
    '''
    print("checkNeighbors.............")
    
    cur_coor = np.column_stack((cur_adata.obs['array_row'].values, cur_adata.obs['array_col'].values))

    # start = time.time()
    if tree_type == 'BallTree':
        cur_coor_tree = BallTree(cur_coor, leaf_size=leaf_size)
    elif tree_type == 'cKDTree':
        cur_coor_tree = cKDTree(cur_coor, leafsize=leaf_size)
    elif tree_type == 'KDTree':
        cur_coor_tree = KDTree(cur_coor, leaf_size=leaf_size)
    else:
        raise ValueError('Invalid tree_type. Only "BallTree, KDTree and cKDTree" is supported.')
    # end = time.time()
    # print(f"Time to build KDTree: {end - start}")

    # start = time.time()
    location_dist, location_ind  = cur_coor_tree.query(cur_coor, k=(neighbor_k+1))
    # end = time.time()
    # print(f"Time to query KDTree: {end - start}")

    # Reshape the results to 2D if they are 1D
    if len(location_dist.shape) == 1:
        location_dist = location_dist[:, None]
    if len(location_ind.shape) == 1:
        location_ind = location_ind[:, None]

    ## Need to consider the selected location itself
    location_dist = location_dist[:,0]
    location_ind = location_ind[:,0]

    ## shumin
    # location_dist = location_dist[:,1:]
    # location_ind = location_ind[:,1:]
    
    return location_dist, location_ind



#################################################################
# 2024.9.16 NameError: name 'loadTrainTestData' is not defined
#################################################################
def loadTrainTestData(train_loader, neighbor_k, tree_type='KDTree', leaf_size=2, dataset_class='Visium'):
    
    tqdm_object = tqdm(train_loader, total=len(train_loader))

    matrix_data = []
    image_data = []
    spatial_coords_list = []
    array_row_list = []
    array_col_list = []

    for batch in tqdm_object:
        # Load data
        matrix_data.append(batch["reduced_expression"].clone().detach().cuda())
        image_data.append(batch["image"].clone().detach().cuda())

        # Process each batch's spatial_coords
        spatial_coords = batch["spatial_coords"]
        combined_coords = torch.stack((spatial_coords[0], spatial_coords[1]), dim=1)
        spatial_coords_list.append(combined_coords)

        array_row = batch["array_row"]
        array_row_list.append(array_row)
        array_col = batch["array_col"]
        array_col_list.append(array_col)

    # Matrix data
    matrix_tensor = torch.cat(matrix_data).to(device)
    # Coord data
    spatial_coords_list_all = torch.cat(spatial_coords_list).to(device)
    array_row_list_all = torch.cat(array_row_list).to(device)
    array_col_list_all = torch.cat(array_col_list).to(device)
    # Image data
    image_tensor = torch.cat(image_data).to(device)
    print('image_tensor:', image_tensor.shape)
    image_tensor = image_tensor.view(image_tensor.shape[0] * image_tensor.shape[1], image_tensor.shape[2])
    inputdata_reshaped, latent_image_reshape = reshape_latent_image(image_tensor, dataset_class)
    latent_representation_image_arr = latent_image_reshape.cpu().detach().numpy()

    # Create adata_latent object
    adata_latent = anndata.AnnData(X=latent_representation_image_arr)
    adata_latent.obsm['spatial'] = np.array(spatial_coords_list_all.cpu())
    adata_latent.obs['array_row'] = np.array(array_row_list_all.cpu())
    adata_latent.obs['array_col'] = np.array(array_col_list_all.cpu())

    train_genes = adata_latent.var_names

    # Generate training data representation, training coordinate matrix, and positive sample information
    cur_train_data_mat = inputdata_reshaped
    cur_train_coors_mat = np.column_stack((adata_latent.obs['array_row'], adata_latent.obs['array_col']))
    cur_train_matrix_mat = matrix_tensor

    # Generate positive pair information
    pos_dist, pos_ind = checkNeighbors(adata_latent, neighbor_k, tree_type, leaf_size)
    cur_pos_info = {'pos dist': pos_dist, 'pos ind': pos_ind}

    return cur_train_data_mat, cur_train_matrix_mat, cur_train_coors_mat, cur_pos_info














"""
FineST: Fine-grained Spatial Transcriptomic
See: https://github.com/LingyuLi-math/FineST
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from pathlib import Path

here = path.abspath(path.dirname(__file__))

# Set __version__ for the project.
exec(open("./FineST/version.py").read())

# Get the long description from the relevant file
long_description=Path("README.rst").read_text("utf-8")
long_description_content_type='text/x-rst'

setup(
    name='FineST',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,

    description='FineST: Fine-grained Spatial Transcriptomic',
    long_description=long_description,
    long_description_content_type='text/x-rst',

    # The project's main homepage.
    url='https://github.com/LingyuLi-math/FineST',

    # Author details
    author=['Lingyu Li', 'Yuanhua Huang'],
    author_email='lingyuli@hku.hk',

    # Choose your license
    license='Apache-2.0',

    # What does your project relate to?
    keywords=[
        'Spatial transcriptomics', 
        'HE stained images',
        'Cell-cell communication', 
        'Ligand-recptor interaction'
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # entry_points={
    #       'console_scripts': [
    #           ],
    #       }, 

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    
    # install_requires=reqs,
    install_requires=[
        l.strip() for l in Path('requirements.txt').read_text('utf-8').splitlines()
    ],

    include_package_data=True,
    package_data={'': ['datasets/*.csv', 'datasets/*.csv.gz']},

    extras_require={
        'docs': [
            #'sphinx == 1.8.3',
            'sphinx_bootstrap_theme']},

    py_modules = ['FineST']

    # buid the distribution: python setup.py sdist
    # upload to pypi: twine upload dist/...
)



from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple, Union, Callable, Optional, Sequence
from inspect import Parameter, signature, Signature
from pathlib import Path
from dataclasses import field, dataclass
import os

from scanpy import read, logging as logg
from anndata import AnnData
from scanpy._utils import check_presence_download
import anndata

PathLike = Union[os.PathLike, str]
Function_t = Callable[..., Union[AnnData, Any]]


@dataclass(frozen=True)  # type: ignore[misc]
class Metadata(ABC):
    """Base class handling metadata."""

    name: str
    url: str

    doc_header: Optional[str] = field(default=None, repr=False)
    path: Optional[PathLike] = field(default=None, repr=False)
    shape: Optional[Tuple[int, int]] = field(default=None, repr=False)
    library_id: Optional[Union[str, Sequence[str]]] = field(default=None, repr=False)

    _DOC_FMT = ""

    def __post_init__(self) -> None:
        if self.doc_header is None:
            object.__setattr__(self, "doc_header", f"Download `{self.name.title().replace('_', ' ')}` data.")
        if self.path is None:
            object.__setattr__(self, "path", os.path.expanduser(f"~/.cache/FineST/{self.name}"))

    @property
    @abstractmethod
    def _extension(self) -> str:
        pass

    @abstractmethod
    def _download(self, fpath: PathLike, backup_url: str, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def _create_signature(self) -> Signature:
        pass

    def _create_function(self, name: str, glob_ns: Dict[str, Any]) -> None:
        # if name in globals():
        #     raise KeyError(f"Function name `{name}` is already present in `{sorted(globals().keys())}`.")

        sig = self._create_signature()
        globals()["NoneType"] = type(None)  # __post_init__ return annotation
        globals()[name] = self

        exec(
            f"def {self.name}{sig}:\n"
            f'    """'
            f"    {self._DOC_FMT.format(doc_header=self.doc_header, shape=self.shape)}"
            f'    """\n'
            f"    return {name}.download(path, **kwargs)".replace(" /,", ""),
            globals(),
            glob_ns,
        )

    def download(self, fpath: Optional[PathLike] = None, **kwargs: Any) -> Any:
        """Download the dataset into ``fpath``."""
        fpath = str(self.path if fpath is None else fpath)
        if not fpath.endswith(self._extension):
            fpath += self._extension

        if os.path.isfile(fpath):
            logg.debug(f"Loading dataset `{self.name}` from `{fpath}`")
        else:
            logg.debug(f"Downloading dataset `{self.name}` from `{self.url}` as `{fpath}`")

        dirname = Path(fpath).parent
        try:
            if not dirname.is_dir():
                logg.info(f"Creating directory `{dirname}`")
                dirname.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logg.error(f"Unable to create directory `{dirname}`. Reason `{e}`")

        data = self._download(fpath=fpath, backup_url=self.url, **kwargs)

        if self.shape is not None and data.shape != self.shape:
            raise ValueError(f"Expected the data to have shape `{self.shape}`, found `{data.shape}`.")

        return data


class AMetadata(Metadata):
    """Metadata class for :class:`anndata.AnnData`."""

    _DOC_FMT = """
    {doc_header}

    The shape of this :class:`anndata.AnnData` object ``{shape}``.

    Parameters
    ----------
    path
        Path where to save the dataset.
    kwargs
        Keyword arguments for :func:`scanpy.read`.

    Returns
    -------
    The dataset."""

    def _create_signature(self) -> Signature:
        return signature(lambda _: _).replace(
            parameters=[
                Parameter("path", kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=PathLike, default=None),
                Parameter("kwargs", kind=Parameter.VAR_KEYWORD, annotation=Any),
            ],
            return_annotation=anndata.AnnData,
        )

    def _download(self, fpath: PathLike, backup_url: str, **kwargs: Any) -> AnnData:
        kwargs.setdefault("sparse", False)
        kwargs.setdefault("cache", True)

        return read(filename=fpath, backup_url=backup_url, **kwargs)

    @property
    def _extension(self) -> str:
        return ".h5ad"


class ImgMetadata(Metadata):
    """Metadata class for :class:`FineST.im.ImageContainer`."""

    _DOC_FMT = """
    {doc_header}

    The shape of this image is ``{shape}``.

    Parameters
    ----------
    path
        Path where to save the .tiff image.
    kwargs
        Keyword arguments for :meth:`FineST.im.ImageContainer.add_img`.

    Returns
    -------
    :class:`FineST.im.ImageContainer`
        The image data."""
    # not the perfect annotation, but better than nothing
    _EXT = ".tiff"

    def _create_signature(self) -> Signature:
        return signature(lambda _: _).replace(
            parameters=[
                Parameter("path", kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=PathLike, default=None),
                Parameter("kwargs", kind=Parameter.VAR_KEYWORD, annotation=Any),
            ],
        )

    def _download(self, fpath: PathLike, backup_url: str, **kwargs: Any) -> Any:
        from FineST.im import ImageContainer  # type: ignore[attr-defined]

        check_presence_download(Path(fpath), backup_url)

        img = ImageContainer()
        img.add_img(fpath, layer="image", library_id=self.library_id, **kwargs)

        return img

    @property
    def _extension(self) -> str:
        return ".tiff"



import numpy as np
import tensorflow as tf
from keras import backend as K


def _nan2zero(x):
    return tf.where(tf.math.is_nan(x), tf.zeros_like(x), x)

def _nan2inf(x):
    return tf.where(tf.math.is_nan(x), tf.zeros_like(x)+np.inf, x)

def _nelem(x):
    nelem = tf.reduce_sum(tf.cast(~tf.math.is_nan(x), tf.float32))
    return tf.cast(tf.where(tf.equal(nelem, 0.), 1., nelem), x.dtype)


def _reduce_mean(x):
    nelem = _nelem(x)
    x = _nan2zero(x)
    return tf.divide(tf.reduce_sum(x), nelem)


def mse_loss(y_true, y_pred):
    ret = tf.square(y_pred - y_true)

    return _reduce_mean(ret)


# In the implementations, I try to keep the function signature
# similar to those of Keras objective functions so that
# later on we can use them in Keras smoothly:
# https://github.com/fchollet/keras/blob/master/keras/objectives.py#L7
def poisson_loss(y_true, y_pred):
    y_pred = tf.cast(y_pred, tf.float32)
    y_true = tf.cast(y_true, tf.float32)

    # we can use the Possion PMF from TensorFlow as well
    # dist = tf.contrib.distributions
    # return -tf.reduce_mean(dist.Poisson(y_pred).log_pmf(y_true))

    nelem = _nelem(y_true)
    y_true = _nan2zero(y_true)

    # last term can be avoided since it doesn't depend on y_pred
    # however keeping it gives a nice lower bound to zero
    ret = y_pred - y_true*tf.math.log(y_pred+1e-10) + tf.math.lgamma(y_true+1.0)

    return tf.divide(tf.reduce_sum(ret), nelem)


# We need a class (or closure) here,
# because it's not possible to
# pass extra arguments to Keras loss functions
# See https://github.com/fchollet/keras/issues/2121

# dispersion (theta) parameter is a scalar by default.
# scale_factor scales the nbinom mean before the
# calculation of the loss to balance the
# learning rates of theta and network weights
class NB(object):
    def __init__(self, theta=None, masking=False, scope='nbinom_loss/',
                 scale_factor=1.0, debug=False):

        # for numerical stability
        self.eps = 1e-10
        self.scale_factor = scale_factor
        self.debug = debug
        self.scope = scope
        self.masking = masking
        self.theta = theta

    def loss(self, y_true, y_pred, mean=True):
        scale_factor = self.scale_factor
        eps = self.eps

        with tf.name_scope(self.scope):
            y_true = tf.cast(y_true, tf.float32)
            y_pred = tf.cast(y_pred, tf.float32) * scale_factor

            if self.masking:
                nelem = _nelem(y_true)
                y_true = _nan2zero(y_true)

            # Clip theta
            theta = tf.minimum(self.theta, 1e6)

            t1 = tf.math.lgamma(theta+eps) + tf.math.lgamma(y_true+1.0) - tf.math.lgamma(y_true+theta+eps)
            t2 = (theta+y_true) * tf.math.log(1.0 + (y_pred/(theta+eps))) + (y_true * (tf.math.log(theta+eps) - tf.math.log(y_pred+eps)))

            if self.debug:
                assert_ops = [
                        tf.verify_tensor_all_finite(y_pred, 'y_pred has inf/nans'),
                        tf.verify_tensor_all_finite(t1, 't1 has inf/nans'),
                        tf.verify_tensor_all_finite(t2, 't2 has inf/nans')]

                tf.summary.histogram('t1', t1)
                tf.summary.histogram('t2', t2)

                with tf.control_dependencies(assert_ops):
                    final = t1 + t2

            else:
                final = t1 + t2

            final = _nan2inf(final)

            if mean:
                if self.masking:
                    final = tf.divide(tf.reduce_sum(final), nelem)
                else:
                    final = tf.reduce_mean(final)


        return final

class ZINB(NB):
    def __init__(self, pi, ridge_lambda=0.0, scope='zinb_loss/', **kwargs):
        super().__init__(scope=scope, **kwargs)
        self.pi = pi
        self.ridge_lambda = ridge_lambda

    def loss(self, y_true, y_pred, mean=True):
        scale_factor = self.scale_factor
        eps = self.eps

        with tf.name_scope(self.scope):
            # reuse existing NB neg.log.lik.
            # mean is always False here, because everything is calculated
            # element-wise. we take the mean only in the end
            nb_case = super().loss(y_true, y_pred, mean=False) - tf.math.log(1.0-self.pi+eps)

            y_true = tf.cast(y_true, tf.float32)
            y_pred = tf.cast(y_pred, tf.float32) * scale_factor
            theta = tf.minimum(self.theta, 1e6)

            zero_nb = tf.pow(theta/(theta+y_pred+eps), theta)
            zero_case = -tf.math.log(self.pi + ((1.0-self.pi)*zero_nb)+eps)
            result = tf.where(tf.less(y_true, 1e-8), zero_case, nb_case)
            ridge = self.ridge_lambda*tf.square(self.pi)
            result += ridge

            if mean:
                if self.masking:
                    result = _reduce_mean(result)
                else:
                    result = tf.reduce_mean(result)

            result = _nan2inf(result)

            if self.debug:
                tf.summary.histogram('nb_case', nb_case)
                tf.summary.histogram('zero_nb', zero_nb)
                tf.summary.histogram('zero_case', zero_case)
                tf.summary.histogram('ridge', ridge)

        return result
